﻿using System;
using System.IO;
using System.Threading;
using System.Windows.Forms;
using System.Diagnostics;
using System.Text;

namespace Alden_USB_CTR_15
{
    public partial class Form1 : Form
    {
        private System.Drawing.Image dotGreen, dotDark;
        public TextBox[] CountsDisplay = new TextBox[30];
        public PictureBox[] Counting = new PictureBox[30];
        private int[] CountingTimeout = new int[30];
        private int[] Counts = new int[30];
        private int[] oldCounts;
        public ushort seconds;
        public Stream logStream;
        private DateTime logStartTime, logEndTime;
        private bool nonNumberEntered = false;
        private static Thread logThread = null;
        private double secondsPerStatusUpdate = 0.1;
        bool logThreadDone = false;
        private uint RunNumber = 1;

        private UInt32[] DeviceIndex = new uint[2];
        private bool bBoard1Found;
        private bool bBoard2Found;

        public Form1()
        {
            InitializeComponent();

            dotGreen = c00.Image;
            dotDark = c01.Image;

            progressBar.Minimum = 0;
            progressBar.Maximum = 1000;

            // J2H -- ToDo: check for 2 configured boards, prompt with instructions / fallback for 1 board
        }

        private void c11_CheckedChanged(object sender, EventArgs e)
        {
            (sender as CheckBox).Text = (sender as CheckBox).Checked ? "Counting" : "";
        }

        private void Form1_Load(object sender, EventArgs e)
        {
            // J2H Todo: verify driver is working, dll is found, etc.
            // J2H Todo: load config file for position and default values for stuff

            Init();
        }

        private void Init()
        {
            timer1.Enabled = false;

            DeviceIndex[0] = AIOUSBNet.AIOUSB.GetDeviceByEEPROMByte(1);
            DeviceIndex[1] = AIOUSBNet.AIOUSB.GetDeviceByEEPROMByte(2);

            bBoard1Found = DeviceIndex[0] != AIOUSBNet.AIOUSB.diNone;
            bBoard2Found = DeviceIndex[1] != AIOUSBNet.AIOUSB.diNone;

            Board1GroupBox.Visible = bBoard1Found;
            Board2GroupBox.Visible = bBoard2Found;

            if (!bBoard1Found) LogText.Text += "Board 1 was not found; ensure Custom EEPROM Offset 0 has been initialized to 1\r\n";
            if (!bBoard2Found) LogText.Text += "Board 2 was not found; ensure Custom EEPROM Offset 0 has been initialized to 2\r\n";

            for (uint i = 0; i < 30; i++)
            {
                Control[] edits = this.Controls.Find("textBox" + (i + 1).ToString(), true); // find all controls called "textBox??" where ?? is 1 through 30
                CountsDisplay[i] = edits[0] as TextBox;
                CountsDisplay[i].Text = "0";

                Control[] checks = this.Controls.Find("c" + i.ToString().PadLeft(2, '0'), true); // find all controls called c?? where ?? is 00 through 29
                Counting[i] = checks[0] as PictureBox;
                Counting[i].Image = dotDark;
                Counting[i].Tag = 0;

                CountingTimeout[i] = 0;
            }

            timer1.Enabled = (bBoard1Found || bBoard2Found) && AutoMonitorCheckbox.Checked;
            StartLogButton.Visible = (bBoard1Found || bBoard2Found);

            { // J2H plausibly improvable
                logPath.Text = Environment.CurrentDirectory;
                if (logPath.Text[logPath.TextLength - 1] != '\\')
                    logPath.Text += '\\';
                logPath.Text += "Alden " + DateTime.Now.ToString("yyyyMMdd HHmmss") + " log.csv";
            }

            if (bBoard1Found)
            {
                AIOUSBNet.AIOUSB.CTR_8254ModeLoad(DeviceIndex[0], 0, (14) /* counter 14 */, 0, (ushort)(seconds - 1)); // seconds duration; counter 14 now outputs a desired duration
                AIOUSBNet.AIOUSB.CTR_8254Mode(DeviceIndex[0], 0, 13, 0); // arm loaded count in 14 by manually issueing one clock out of 13
                AIOUSBNet.AIOUSB.CTR_8254Mode(DeviceIndex[0], 0, 13, 1);

                for (uint i = 0; i < 12; i++)
                {
                    var status = AIOUSBNet.AIOUSB.CTR_8254ModeLoad(DeviceIndex[0], 0, i, 2, 0);
                    if (status != 0) LogText.Text += "buttonLog status: " + status.ToString() + "\r\n";
                }
            }

            if (bBoard2Found)
            {
                AIOUSBNet.AIOUSB.CTR_8254ModeLoad(DeviceIndex[1], 0, (14) /* counter 14 */, 0, (ushort)(seconds - 1)); // seconds duration; counter 14 now outputs a desired duration
                AIOUSBNet.AIOUSB.CTR_8254Mode(DeviceIndex[1], 0, 13, 0); // arm loaded count in 14 by manually issueing one clock out of 13
                AIOUSBNet.AIOUSB.CTR_8254Mode(DeviceIndex[1], 0, 13, 1);

                for (uint i = 0; i < 12; i++)
                {
                    var status = AIOUSBNet.AIOUSB.CTR_8254ModeLoad(DeviceIndex[1], 0, i, 2, 0);
                    if (status != 0) LogText.Text += "buttonLog status: " + status.ToString() + "\r\n";
                }
            }

            if (bBoard1Found)
            {
                AIOUSBNet.AIOUSB.CTR_8254ModeLoad(DeviceIndex[0], 0, (13) /* counter 13 */, 2, 10000); // divide-by-1000; counter 13 now outputs at 1Hz
                AIOUSBNet.AIOUSB.CTR_8254ModeLoad(DeviceIndex[0], 0, (12) /* counter 12 */, 2, 1000);  // divide-by-10000; counter 12 now outputs at 1kHz
                                                                                                       // output of (14) needs to be tied __through inverter__ to GATEs of counters (3-14)
            }

            if (bBoard2Found)
            {
                AIOUSBNet.AIOUSB.CTR_8254ModeLoad(DeviceIndex[1], 0, (13) /* counter 13 */, 2, 10000); // divide-by-1000; counter 13 now outputs at 1Hz
                AIOUSBNet.AIOUSB.CTR_8254ModeLoad(DeviceIndex[1], 0, (12) /* counter 12 */, 2, 1000);  // divide-by-10000; counter 12 now outputs at 1kHz
                                                                                                       // output of (14) needs to be tied __through inverter__ to GATEs of counters (3-14)
            }
        }

        private void Tick()
        {
            int[] counts = new int[CountsDisplay.Length];

            byte[] buf1 = new byte[45];
            byte[] buf2 = new byte[45];
            UInt32 buflen = 45;

            if (bBoard1Found && AutoMonitorCheckbox.Checked)
            {
                var status = AIOUSBNet.AIOUSB.GenericVendorRead(DeviceIndex[0], 0x25, 0, 0, ref buflen, buf1);
                if (status != 0)
                    LogText.Text += "Tick status: " + status.ToString() + "\r\n";
                if (buflen != 45)
                    LogText.Text += "genread status len == " + buflen.ToString() + "\r\n  Are you using old firmware?\r\n";
            }

            if (bBoard2Found && AutoMonitorCheckbox.Checked)
            {
                var status = AIOUSBNet.AIOUSB.GenericVendorRead(DeviceIndex[1], 0x25, 0, 0, ref buflen, buf2);
                if (status != 0)
                    LogText.Text += "Tick status: " + status.ToString() + "\r\n";
                if (buflen != 45)
                    LogText.Text += "genread status len == " + buflen.ToString() + "\r\n  Are you using old firmware?\r\n";
            }

            if (oldCounts == null)
            {
                oldCounts = counts;
                return;
            }

            if (bBoard1Found)
            {
                for (uint i = 0; i < 15; i++)
                {
                    if ((buf1[2 + i * 3] & 0b01000000) == 0)
                        counts[i] = (65536 - (buf1[i * 3 + 1] * 256 + buf1[i * 3]) + 1) & 0xFFFF; // if not NULL COUNT
                    else
                        counts[i] = 0;

                }
            }

            if (bBoard2Found)
            {
                for (uint i = 0; i < 15; i++)
                {
                    if ((buf2[2 + i * 3] & 0b01000000) == 0) //  was buf1[]
                        counts[i + 15] = (65536 - (buf2[i * 3 + 1] * 256 + buf2[i * 3]) + 1) & 0xFFFF; // if not NULL COUNT
                    else
                        counts[i + 15] = 0;

                }
            }

            for (int i = 0; i < 30; i++)
            {
                if (oldCounts[i] != counts[i])
                {
                    Counting[i].Image = dotGreen;
                    CountsDisplay[i].Text = counts[i].ToString();

                    CountingTimeout[i] = 10;
                }
                else if (CountingTimeout[i] < 0) { }
                else if (CountingTimeout[i] == 0)
                {
                    Counting[i].Image = dotDark;
                    CountingTimeout[i] = -1;
                    CountsDisplay[i].Text = "";
                }
                else
                {
                    CountingTimeout[i]--;
                }
            }

            oldCounts = counts;
        }

        private void timer1_Tick(object sender, EventArgs e)
        {
            Tick();
        }

        private void LogFileBrowse(object sender, EventArgs e)
        {
            saveFileDialog1.FileName = Path.GetFileName(logPath.Text);
            saveFileDialog1.InitialDirectory = Path.GetDirectoryName(logPath.Text);
            saveFileDialog1.Filter = "csv files (*.csv)|*.csv";
            saveFileDialog1.FilterIndex = 1;
            saveFileDialog1.RestoreDirectory = true;
            if (saveFileDialog1.ShowDialog() == DialogResult.OK)
                logPath.Text = saveFileDialog1.FileName;
        }

        private void AutoMonitorCheckbox_CheckedChanged(object sender, EventArgs e)
        {
            if (AutoMonitorCheckbox.Checked)
            {
                if (bBoard1Found) AIOUSBNet.AIOUSB.CTR_8254Mode(DeviceIndex[0], 0, 14, 0); // free-run GATE to counters when not logging
                if (bBoard2Found) AIOUSBNet.AIOUSB.CTR_8254Mode(DeviceIndex[1], 0, 14, 0); // free-run GATE to counters when not logging
                timer1.Enabled = true;
            }
            else
            {
                timer1.Enabled = false;
                if (bBoard1Found) AIOUSBNet.AIOUSB.CTR_8254Mode(DeviceIndex[0], 0, 14, 1); // disable gates
                if (bBoard2Found) AIOUSBNet.AIOUSB.CTR_8254Mode(DeviceIndex[1], 0, 14, 1); // disable gates
            }
        }
        private void StartLogButton_Click(object sender, EventArgs e)
        {
            bool bTRYAGAIN = false;
            bool bCancel = false;
            do
            {
                
                try
                {
                    logStream = File.Open(logPath.Text, FileMode.Append, FileAccess.Write, FileShare.Read);
                }
                catch (Exception ex)
                {
                    DialogResult result = MessageBox.Show("The log file did not open for write; is it open already?", "Cannot Open Log File", MessageBoxButtons.RetryCancel);
                    if (result == System.Windows.Forms.DialogResult.Cancel)
                    {
                        
                        bCancel = true;
                        bTRYAGAIN = false;
                    }
                    else
                        bTRYAGAIN = true;
                }
            } while (bTRYAGAIN);

            if (bCancel)
                return;

            logThreadDone = false;
            StartLogButton.Enabled = false;
//            AutoMonitorCheckbox.Visible = false;
            AbortButton.Visible = true;
            ResetCounterButton.Visible = false;

            progressBar.Value = 0;
            progressBar.Visible = true;

            ushort.TryParse(logSeconds.Text, out seconds);

            if (bBoard1Found)
            {
                AIOUSBNet.AIOUSB.CTR_8254ModeLoad(DeviceIndex[0], 0, (14) /* counter 14 */, 0, (ushort)(seconds - 1)); // seconds duration; counter 14 now outputs a desired duration
                AIOUSBNet.AIOUSB.CTR_8254Mode(DeviceIndex[0], 0, 13, 0); // arm loaded count in 14 by manually issueing one clock out of 13
                AIOUSBNet.AIOUSB.CTR_8254Mode(DeviceIndex[0], 0, 13, 1);

                for (uint i = 0; i < 12; i++)
                {
                    var status = AIOUSBNet.AIOUSB.CTR_8254ModeLoad(DeviceIndex[0], 0, i, 2, 0);
                    if (status != 0) LogText.Text += "buttonLog status: " + status.ToString() + "\r\n";
                }
            }

            if (bBoard2Found)
            {
                AIOUSBNet.AIOUSB.CTR_8254ModeLoad(DeviceIndex[1], 0, (14) /* counter 14 */, 0, (ushort)(seconds - 1)); // seconds duration; counter 14 now outputs a desired duration
                AIOUSBNet.AIOUSB.CTR_8254Mode(DeviceIndex[1], 0, 13, 0); // arm loaded count in 14 by manually issueing one clock out of 13
                AIOUSBNet.AIOUSB.CTR_8254Mode(DeviceIndex[1], 0, 13, 1);

                for (uint i = 0; i < 12; i++)
                {
                    var status = AIOUSBNet.AIOUSB.CTR_8254ModeLoad(DeviceIndex[1], 0, i, 2, 0);
                    if (status != 0) LogText.Text += "buttonLog status: " + status.ToString() + "\r\n";
                }
            }

            if (bBoard1Found)
            {
                AIOUSBNet.AIOUSB.CTR_8254ModeLoad(DeviceIndex[0], 0, (13) /* counter 13 */, 2, 10000); // divide-by-1000; counter 13 now outputs at 1Hz
                AIOUSBNet.AIOUSB.CTR_8254ModeLoad(DeviceIndex[0], 0, (12) /* counter 12 */, 2, 1000);  // divide-by-10000; counter 12 now outputs at 1kHz
                                                                                                       // output of (14) needs to be tied __through inverter__ to GATEs of counters (3-14)
            }

            if (bBoard2Found)
            {
                AIOUSBNet.AIOUSB.CTR_8254ModeLoad(DeviceIndex[1], 0, (13) /* counter 13 */, 2, 10000); // divide-by-1000; counter 13 now outputs at 1Hz
                AIOUSBNet.AIOUSB.CTR_8254ModeLoad(DeviceIndex[1], 0, (12) /* counter 12 */, 2, 1000);  // divide-by-10000; counter 12 now outputs at 1kHz
                                                                                                       // output of (14) needs to be tied __through inverter__ to GATEs of counters (3-14)
            }

            logThread = new Thread(new ThreadStart(DoLogThread))
            {
                Priority = ThreadPriority.Highest // J2H TODO: make config option
            };
            logThread.Start();
        }

        private void displayLogStatus(double progress)
        {
            progressBar.Value = Math.Max(0, Math.Min(Convert.ToInt32(progress * progressBar.Maximum), progressBar.Maximum));
        }

        private void DoLogThread()
        {
            logStartTime = DateTime.Now;

            long beginTime = Stopwatch.GetTimestamp();
            long endTime = beginTime + seconds * Stopwatch.Frequency;
            long ticksPerStatusUpdate = Convert.ToInt64(Stopwatch.Frequency * secondsPerStatusUpdate);
            long nextStatusUpdateTime = beginTime + ticksPerStatusUpdate;
            byte[] buf0 = new byte[45];
            byte[] buf1 = new byte[45];
            UInt32 buflen = 45;

            do
            {
                long now = Stopwatch.GetTimestamp();
                if (now >= nextStatusUpdateTime)
                {
                    nextStatusUpdateTime += ticksPerStatusUpdate;
                    {
                        UInt16 v0 = 0, v1 = 0;
                        byte s0 = 0xBF, s1 = 0xBF;

                        if (bBoard1Found) AIOUSBNet.AIOUSB.CTR_8254ReadStatus(DeviceIndex[0], 0, 14, ref v0, ref s0); // check for NULL COUNT, board 0
                        if (bBoard2Found) AIOUSBNet.AIOUSB.CTR_8254ReadStatus(DeviceIndex[1], 0, 14, ref v1, ref s1); // check for NULL COUNT

                        if (((s0 | s1) & 0x40) == 0) // if not NULL count (on _both_ installed boards)
                            Invoke((MethodInvoker)delegate { displayLogStatus((seconds - (v0 * 1.0)) / seconds); });

                        if (((s0 & s1) & 0x80) != 0) // OUT pin is HIGH (on _both_ installed boards)
                            logThreadDone = true;
                    }
                }
            } while (!logThreadDone);

            if (bBoard1Found)
            {
                var status = AIOUSBNet.AIOUSB.GenericVendorRead(DeviceIndex[0], 0x25, 0, 0, ref buflen, buf0); // statusAll
                if (status != 0) Invoke((MethodInvoker)delegate { LogText.Text += "Tick0 status: " + status.ToString() + "\r\n"; });
                if (buflen != 45) Invoke((MethodInvoker)delegate { LogText.Text += "genread0 status len == " + buflen.ToString() + "\r\n"; });
            }

            if (bBoard2Found)
            {
                var status = AIOUSBNet.AIOUSB.GenericVendorRead(DeviceIndex[1], 0x25, 0, 0, ref buflen, buf1); // statusAll
                if (status != 0) Invoke((MethodInvoker)delegate { LogText.Text += "Tick1 status: " + status.ToString() + "\r\n"; });
                if (buflen != 45) Invoke((MethodInvoker)delegate { LogText.Text += "genread1 status len == " + buflen.ToString() + "\r\n"; });
            }

            if (bBoard1Found)
                for (uint i = 0; i < 15; i++)
                {
                    if ((buf0[2 + i * 3] & 0b01000000) == 0)
                        Counts[i] = (65536 - (buf0[i * 3 + 1] * 256 + buf0[i * 3]) + 1) & 0xFFFF; // if not NULL COUNT
                    else
                        Counts[i] = 0;
                }

            if (bBoard2Found)
                for (uint i = 0; i < 15; i++)
                {
                    if ((buf1[2 + i * 3] & 0b01000000) == 0)
                        Counts[i + 15] = (65536 - (buf1[i * 3 + 1] * 256 + buf1[i * 3]) + 1) & 0xFFFF; // if not NULL COUNT
                    else
                        Counts[i + 15] = 0;
                }

            logEndTime = DateTime.Now;

            Invoke((MethodInvoker)delegate { CloseLog(); });
        }

        private void CloseLog()
        {
            progressBar.Value = progressBar.Maximum;

            String LogLine;

            if (RunNumber == 1)
            {
                LogLine = "\"Date\",\"Log Start Time\",\"Run #\",\"Duration\",\"CH0\",\"CH1\",\"CH2\",\"CH3\",\"CH4\",\"CH5\",\"CH6\",\"CH7\",\"CH8\",\"CH9\",\"CH10\",\"CH11\",\"CH12\",\"CH13\",\"CH14\",\"CH15\",\"CH16\",\"CH17\",\"CH18\",\"CH19\",\"CH20\",\"CH21\",\"CH22\",\"CH23\",\"CH24\",\"CH25\",\"CH26\",\"CH27\",\"CH28\",\"CH29\",\"Note\"\r\n";
                byte[] LogHeaderBytes = new UTF8Encoding(true).GetBytes(LogLine);

                logStream.Write(LogHeaderBytes, 0, LogHeaderBytes.Length);
            }

            LogLine = DateTime.Now.ToString("M/dd/yyyy") + "," + DateTime.Now.ToString("HH:mm:ss") + "," + RunNumber.ToString() + "," + seconds.ToString();
            for (int i = 0; i < 30; i++)
                LogLine += "," + Counts[i].ToString()
            ;
            LogLine += ",\"" + NoteTextBox.Text + "\"";
            LogLine += "\r\n";
            byte[] LogBytes = new UTF8Encoding(true).GetBytes(LogLine);

            logStream.Write(LogBytes, 0, LogBytes.Length);
            logStream.Close();

            ++RunNumber;
            RunLogNoteLabel.Text = "Log note for Run: " + RunNumber.ToString();

            StartLogButton.Enabled = true;
            //AutoMonitorCheckbox.Visible = true;
            ResetCounterButton.Visible = true;
            AbortButton.Visible = false;

            if (AutoMonitorCheckbox.Checked)
            {
                if (bBoard1Found) AIOUSBNet.AIOUSB.CTR_8254Mode(DeviceIndex[0], 0, 14, 0); // free-run GATE to counters when not logging
                if (bBoard2Found) AIOUSBNet.AIOUSB.CTR_8254Mode(DeviceIndex[1], 0, 14, 0); // free-run GATE to counters when not logging
            }
            progressBar.Visible = false;
        }
        private void textBox17_KeyDown(object sender, KeyEventArgs e)
        {
            nonNumberEntered = false;
            if (e.KeyCode < Keys.D0 || e.KeyCode > Keys.D9)
                if (e.KeyCode < Keys.NumPad0 || e.KeyCode > Keys.NumPad9)
                    if (e.KeyCode != Keys.Back)
                        nonNumberEntered = true;

            if (Control.ModifierKeys == Keys.Shift)
                nonNumberEntered = true;
        }

        private void textBox17_KeyPress(object sender, KeyPressEventArgs e)
        {
            if (nonNumberEntered == true)
                e.Handled = true;
        }

        private void ResetCounterButton_Click(object sender, EventArgs e)
        {
            if (bBoard1Found)
            {
                AIOUSBNet.AIOUSB.CTR_8254ModeLoad(DeviceIndex[0], 0, (14) /* counter 14 */, 0, (ushort)(seconds - 1)); // seconds duration; counter 14 now outputs a desired duration
                AIOUSBNet.AIOUSB.CTR_8254Mode(DeviceIndex[0], 0, 13, 0); // arm loaded count in 14 by manually issueing one clock out of 13
                AIOUSBNet.AIOUSB.CTR_8254Mode(DeviceIndex[0], 0, 13, 1);

                for (uint i = 0; i < 12; i++)
                {
                    var status = AIOUSBNet.AIOUSB.CTR_8254ModeLoad(DeviceIndex[0], 0, i, 2, 0);
                    if (status != 0) LogText.Text += "buttonLog status: " + status.ToString() + "\r\n";
                }
            }

            if (bBoard2Found)
            {
                AIOUSBNet.AIOUSB.CTR_8254ModeLoad(DeviceIndex[1], 0, (14) /* counter 14 */, 0, (ushort)(seconds - 1)); // seconds duration; counter 14 now outputs a desired duration
                AIOUSBNet.AIOUSB.CTR_8254Mode(DeviceIndex[1], 0, 13, 0); // arm loaded count in 14 by manually issueing one clock out of 13
                AIOUSBNet.AIOUSB.CTR_8254Mode(DeviceIndex[1], 0, 13, 1);

                for (uint i = 0; i < 12; i++)
                {
                    var status = AIOUSBNet.AIOUSB.CTR_8254ModeLoad(DeviceIndex[1], 0, i, 2, 0);
                    if (status != 0) LogText.Text += "buttonLog status: " + status.ToString() + "\r\n";
                }
            }

            if (bBoard1Found)
            {
                AIOUSBNet.AIOUSB.CTR_8254ModeLoad(DeviceIndex[0], 0, (13) /* counter 13 */, 2, 10000); // divide-by-1000; counter 13 now outputs at 1Hz
                AIOUSBNet.AIOUSB.CTR_8254ModeLoad(DeviceIndex[0], 0, (12) /* counter 12 */, 2, 1000);  // divide-by-10000; counter 12 now outputs at 1kHz
                                                                                                       // output of (14) needs to be tied __through inverter__ to GATEs of counters (3-14)
            }

            if (bBoard2Found)
            {
                AIOUSBNet.AIOUSB.CTR_8254ModeLoad(DeviceIndex[1], 0, (13) /* counter 13 */, 2, 10000); // divide-by-1000; counter 13 now outputs at 1Hz
                AIOUSBNet.AIOUSB.CTR_8254ModeLoad(DeviceIndex[1], 0, (12) /* counter 12 */, 2, 1000);  // divide-by-10000; counter 12 now outputs at 1kHz
                                                                                                       // output of (14) needs to be tied __through inverter__ to GATEs of counters (3-14)
            }
        }

        private void Form1_FormClosing(object sender, FormClosingEventArgs e)
        {
            logThreadDone = true;
            if (logThread != null)
                while (logThread.IsAlive)
                    Application.DoEvents();

            Properties.Settings.Default.Save();
        }

        private void AbortButton_Click(object sender, EventArgs e)
        {
            logThreadDone = true;
        }
    }
}
